# -*- coding: utf-8 -*-
"""
Create, Read, Update, Delete functionality

"""

from sqlalchemy import cast, Text, insert, update, delete, select
from typing import List
from geoalchemy2.types import Geometry
from geoalchemy2.elements import WKBElement, WKTElement
from sqlalchemy.orm import Session
from sqlalchemy.engine import Engine
from sqlalchemy.dialects.postgresql import JSONB, ARRAY
from fastapi import File, UploadFile
from fastapi.responses import JSONResponse
from . import models
from .models import StationmetaCore, StationmetaChangelog, stationmeta_core_stationmeta_roles_table, \
                    stationmeta_core_stationmeta_annotations_table, \
                    CZ_enum, CV_enum, ST_enum, TA_enum, TC_enum, CN_enum, TZ_enum
from toardb.generic.models import RS_enum, RC_enum, AK_enum, CL_enum
from .schemas import get_coordinates_from_geom, get_geom_from_coordinates, StationmetaCreate, StationmetaPatch, Coordinates
from pydantic import ValidationError
from toardb.utils.utils import get_value_from_str, get_str_from_value


def get_stationmeta_core(db: Session, station_code: str):
    db_object = db.query(models.StationmetaCore).filter(models.StationmetaCore.codes.any(station_code.strip())).first()
    # there is a mismatch with coordinates and additional_metadata
    if db_object:
        if isinstance(db_object.coordinates, (WKBElement, WKTElement)):
            db_object.coordinates = get_coordinates_from_geom(db_object.coordinates)
        db_object.additional_metadata = str(db_object.additional_metadata).replace("'",'"')
    return db_object


def get_stationmeta(db: Session, station_code: str):
    db_object = db.query(models.StationmetaCore).filter(models.StationmetaCore.codes.any(station_code.strip())).first()
    # there is a mismatch with coordinates and additional_metadata
    if db_object:
        if isinstance(db_object.coordinates, (WKBElement, WKTElement)):
            db_object.coordinates = get_coordinates_from_geom(db_object.coordinates)
        db_object.additional_metadata = str(db_object.additional_metadata).replace("'",'"')
    return db_object


def get_stationmeta_by_id(db: Session, station_id: int):
    db_object = db.query(models.StationmetaCore).filter(models.StationmetaCore.id == station_id).first()
    # there is a mismatch with coordinates and additional_metadata
    if db_object:
        if isinstance(db_object.coordinates, (WKBElement, WKTElement)):
            db_object.coordinates = get_coordinates_from_geom(db_object.coordinates)
        db_object.additional_metadata = str(db_object.additional_metadata).replace("'",'"')
    return db_object


def get_all_stationmeta_core(db: Session, skip : int = 0, limit: int = None):
    db_objects = db.query(models.StationmetaCore).offset(skip).limit(limit).all()
    for db_object in db_objects:
        # there is a mismatch with coordinates and additional_metadata
        if isinstance(db_object.coordinates, (WKBElement, WKTElement)):
            db_object.coordinates = get_coordinates_from_geom(db_object.coordinates)
        db_object.additional_metadata = str(db_object.additional_metadata).replace("'",'"')
    return db_objects


def get_all_stationmeta(db: Session, skip : int = 0, limit: int = None):
    db_objects = db.query(models.StationmetaCore).offset(skip).limit(limit).all()
    for db_object in db_objects:
        # there is a mismatch with coordinates and additional_metadata
        if isinstance(db_object.coordinates, (WKBElement, WKTElement)):
            db_object.coordinates = get_coordinates_from_geom(db_object.coordinates)
        db_object.additional_metadata = str(db_object.additional_metadata).replace("'",'"')
    return db_objects


# is this internal, or should this also go to public REST api?
# - get_stationmeta_roles
# - get_stationmeta_changelog
# - get_unique_stationmeta_role
# - get_stationmeta_role_by_id
# - get_unique_stationmeta_annotation
def get_stationmeta_roles(db: Session, station_id: int):
    return db.execute(select([stationmeta_core_stationmeta_roles_table]).where(stationmeta_core_stationmeta_roles_table.c.station_id == station_id))


def get_stationmeta_changelog(db: Session, station_id: int):
    return db.query(models.StationmetaChangelog).filter(models.StationmetaChangelog.station_id == station_id).all()


def get_unique_stationmeta_role(db: Session, role: int, contact_id: int, status: int):
    db_object = db.query(models.StationmetaRole).filter(models.StationmetaRole.role == role) \
                                                .filter(models.StationmetaRole.contact_id == contact_id) \
                                                .filter(models.StationmetaRole.status == status) \
                                                .first()
    return db_object

def get_stationmeta_role_by_id(db: Session, role_id):
    return db.query(models.StationmetaRole).filter(models.StationmetaRole.id == role_id).first()


def get_unique_stationmeta_annotation(db: Session, text: str, contributor_id: int):
    db_object = db.query(models.StationmetaAnnotation).filter(models.StationmetaAnnotation.text == text) \
                                                      .filter(models.StationmetaAnnotation.contributor_id == contributor_id) \
                                                      .first()
    return db_object


# this is just to fake what would be done! This is a dry-run!
def create_stationmeta_dryrun(db: Session, engine: Engine, stationmeta: StationmetaCreate):
    stationmeta_dict = stationmeta.dict()
    roles_data         = stationmeta_dict.pop('roles', None)
    annotations_data   = stationmeta_dict.pop('annotations', None)
    aux_images_data    = stationmeta_dict.pop('aux_images', None)
    aux_docs_data      = stationmeta_dict.pop('aux_docs', None)
    aux_urls_data      = stationmeta_dict.pop('aux_urls', None)
    globalmeta_data    = stationmeta_dict.pop('globalmeta', None)
    globalservice_data = stationmeta_dict.pop('globalservice', None)
    db_stationmeta = models.StationmetaCore(**stationmeta_dict)
    # there's a mismatch with coordinates --> how to automatically switch back and forth?!
    tmp_coordinates = db_stationmeta.coordinates
    db_stationmeta.coordinates = get_geom_from_coordinates(Coordinates(**db_stationmeta.coordinates))
    # there's also a mismatch with additional_metadata --> BUT: this should not be switched back!
    # in upload command, we have now: "additional_metadata": "{}"
    # but return from this method gives: "additional_metadata": {}
    # ==> there is a mismatch between model(JSONB) and schema(JSON)
    db_stationmeta.additional_metadata = str(db_stationmeta.additional_metadata).replace("'",'"')
    fake_conn = engine.raw_connection()
    fake_cur = fake_conn.cursor()
    radius = 10
    db_cmd = f"select * from stationmeta_core where ST_DistanceSphere(stationmeta_core.coordinates, '{db_stationmeta.coordinates}') < {radius}"
    fake_cur.execute(db_cmd)
    records = fake_cur.fetchall()
    if records:
        if len(records) == 1:
            message=f'already in database --> patch record (add station code)'
        else:
            message = 'more than one station falls within the given radius!\n' + \
                      'choose which station record to patch (add station code)'
        for record in records: 
            message += str(record)
    else:
        message=f'new station: {db_stationmeta.codes},{db_stationmeta.name},{db_stationmeta.coordinates}'
    status_code=200
    return JSONResponse(status_code=status_code, content=message)

def create_stationmeta(db: Session, engine: Engine, stationmeta: StationmetaCreate):
    stationmeta_dict = stationmeta.dict()
    roles_data         = stationmeta_dict.pop('roles', None)
    annotations_data   = stationmeta_dict.pop('annotations', None)
    aux_images_data    = stationmeta_dict.pop('aux_images', None)
    aux_docs_data      = stationmeta_dict.pop('aux_docs', None)
    aux_urls_data      = stationmeta_dict.pop('aux_urls', None)
    globalmeta_data    = stationmeta_dict.pop('globalmeta', None)
    globalservice_data = stationmeta_dict.pop('globalservice', None)
    db_stationmeta = models.StationmetaCore(**stationmeta_dict)
    # there's a mismatch with coordinates --> how to automatically switch back and forth?!
    tmp_coordinates = db_stationmeta.coordinates
    db_stationmeta.coordinates = get_geom_from_coordinates(Coordinates(**db_stationmeta.coordinates))
    # there's also a mismatch with additional_metadata --> BUT: this should not be switched back!
    # in upload command, we have now: "additional_metadata": "{}"
    # but return from this method gives: "additional_metadata": {}
    # ==> there is a mismatch between model(JSONB) and schema(JSON)
    db_stationmeta.additional_metadata = str(db_stationmeta.additional_metadata).replace("'",'"')
    fake_conn = engine.raw_connection()
    fake_cur = fake_conn.cursor()
    radius = 10
    db_cmd = f"select * from stationmeta_core where ST_DistanceSphere(stationmeta_core.coordinates, '{db_stationmeta.coordinates}') < {radius}"
    fake_cur.execute(db_cmd)
    records = fake_cur.fetchall()
    if records:
        if len(records) == 1:
            print('already in database --> patch record (add station code)')
        else:
            print('more than one station falls within the given radius!')
            print('choose which station record to patch (add station code)')
        for record in records: 
            print(record)
    else:
        db_stationmeta.country = get_value_from_str(CN_enum,db_stationmeta.country)
        db_stationmeta.timezone = get_value_from_str(TZ_enum,db_stationmeta.timezone)
        db_stationmeta.type_of_environment = get_value_from_str(ST_enum,db_stationmeta.type_of_environment)
        db_stationmeta.type_of_area = get_value_from_str(TA_enum,db_stationmeta.type_of_area)
        db_stationmeta.coordinate_validation_status = get_value_from_str(CV_enum,db_stationmeta.coordinate_validation_status)
        db.add(db_stationmeta)
        result = db.commit()
        db.refresh(db_stationmeta)
        # get stationmeta_core_id
        stationmeta_core_id = db_stationmeta.id
        # store roles and update association table
        if roles_data:
            for r in roles_data:
                db_role = models.StationmetaRole(**r)
                db_role.role = get_value_from_str(RC_enum,db_role.role)
                db_role.status = get_value_from_str(RS_enum,db_role.status)
                # check whether role is already present in database
                db_object = get_unique_stationmeta_role(db, db_role.role, db_role.contact_id, db_role.status)
                if db_object:
                    role_id = db_object.id
                else:
                    db.add(db_role)
                    db.commit()
                    db.refresh(db_role)
                    role_id = db_role.id
                db.execute(insert(stationmeta_core_stationmeta_roles_table).values(station_id=stationmeta_core_id, role_id=role_id))
                db.commit()
        # store annotations and update association table
        if annotations_data:
            for a in annotations_data:
                db_annotation = models.StationmetaAnnotation(**a)
                # check whether annotation is already present in database
                db_object = get_unique_stationmeta_annotation(db, db_annotation.text, db_annotation.contributor_id)
                if db_object:
                    annotation_id = db_object.id
                else:
                    db.add(db_annotation)
                    db.commit()
                    db.refresh(db_annotation)
                    annotation_id = db_annotation.id
                db.execute(insert(stationmeta_core_stationmeta_annotations_table).values(station_id=stationmeta_core_id, annotation_id=annotation_id))
                db.commit()
        # store aux_images
        if aux_images_data:
            for i in aux_images_data:
                db_aux_image = models.StationmetaAuxImage(**i)
                db_aux_image.station_id = stationmeta_core_id
                db.add(db_aux_image)
                db.commit()
                db.refresh(db_aux_image)
        # store aux_docs
        if aux_docs_data:
            for d in aux_docs_data:
                db_aux_doc = models.StationmetaAuxDoc(**d)
                db_aux_doc.station_id = stationmeta_core_id
                db.add(db_aux_doc)
                db.commit()
                db.refresh(db_aux_doc)
        # store aux_urls
        if aux_urls_data:
            for u in aux_urls_data:
                db_aux_url = models.StationmetaAuxUrl(**u)
                db_aux_url.station_id = stationmeta_core_id
                db.add(db_aux_url)
                db.commit()
                db.refresh(db_aux_url)
        # store globalmeta
        if globalmeta_data:
            db_global = models.StationmetaGlobal(**globalmeta_data)
            if db_global.climatic_zone:
                db_global.climatic_zone = get_value_from_str(CZ_enum,db_global.climatic_zone)
            if db_global.toar1_category:
                db_global.toar1_category = get_value_from_str(TC_enum,db_global.toar1_category)
            db_global.station_id = stationmeta_core_id
            db.add(db_global)
            db.commit()
            db.refresh(db_global)
        # store globalservice
        if globalservice_data:
            db_globalservice = models.StationmetaGlobalService(**globalservice)
            db_globalservice.station_id = stationmeta_core_id
            db.add(db_globalservice)
            db.commit()
            db.refresh(db_globalservice)
    # there's a mismatch with coordinates --> how to automatically switch back and forth?!
    db_stationmeta.coordinates = tmp_coordinates
    return db_stationmeta

def get_field_from_record(db, station_id, field, db_stationmeta):
    if field in ['roles', 'annotations', 'aux_images', 'aux_docs', 'aux_urls', 'globalmeta']:
        field =  f"'{field}': " + str(getattr(db_stationmeta,field))
    else:
        field =  f"'{field}': " + str(getattr(db_stationmeta,field))
    return field

def patch_stationmeta(db: Session, description: str, station_id: int, stationmeta: StationmetaPatch):
    stationmeta_dict       = stationmeta.dict()
    # delete empty fields from stationmeta_dict already at this place, to be able to
    # distinguish between "single value correction in metadata" and "comprehensive metadata revision"
    # (see controlled vocabulary "CL_vocabulary")
    stationmeta_dict2 = {k: v for k, v in stationmeta_dict.items() if v is not None}
    number_of_elements = len(stationmeta_dict2)
    if (number_of_elements == 1):
        type_of_change = get_value_from_str(CL_enum,"SingleValue")
    else:
        type_of_change = get_value_from_str(CL_enum,"Comprehensive")
    roles_data         = stationmeta_dict.pop('roles', None)
    annotations_data   = stationmeta_dict.pop('annotations', None)
    aux_images_data    = stationmeta_dict.pop('aux_images', None)
    aux_docs_data      = stationmeta_dict.pop('aux_docs', None)
    aux_urls_data      = stationmeta_dict.pop('aux_urls', None)
    globalmeta_data    = stationmeta_dict.pop('globalmeta', None)
    globalservice_data = stationmeta_dict.pop('globalservice', None)
    # there's a mismatch with coordinates --> how to automatically switch back and forth?!
    # ==> the following two commands are not working
    # ==> workaround
    # db.query(models.StationmetaCore).filter(models.StationmetaCore.id == stationm_id).update(stationmeta_dict2)
    db_obj = models.StationmetaCore(**stationmeta_dict)
    tmp_coordinates = db_obj.coordinates
    db_stationmeta = db.query(models.StationmetaCore).get(station_id)
    db_stationmeta.coordinates = get_geom_from_coordinates(db_stationmeta.coordinates)
    old_value="{"
    number_of_commas = number_of_elements - 1
    for k, v in stationmeta_dict.items():
        if v is not None:
            # prepare changelog entry
            if (number_of_commas == 0):
                old_value=old_value + get_field_from_record(db, station_id, k, db_stationmeta) + '}'
            else:
                old_value=old_value + get_field_from_record(db, station_id, k, db_stationmeta) + ','
            number_of_commas = number_of_commas - 1
            setattr(db_stationmeta,k,stationmeta_dict[k])
    result = db.commit()
    db.refresh(db_stationmeta)
    # store roles and update association table
    if roles_data:
        # prepare changelog entry/entries
        description = description + f"; add role"
        db_old_roles = get_stationmeta_roles(db, station_id)
        old_value = old_value + "'roles': {"
        for oldr in db_old_roles:
            old_role = get_stationmeta_role_by_id(db, oldr.role_id)
            old_value=old_value + "{'role': '" + get_str_from_value(RC_enum,old_role.role) + \
                                "', 'status': '" + get_str_from_value(RS_enum,old_role.status) + \
                                "', 'contact_id': " + str(old_role.contact_id) + '}'
        if (number_of_commas == 0):
            old_value=old_value + '}'
        else:
            old_value=old_value + ','
        number_of_commas = number_of_commas - 1
        for r in roles_data:
            db_role = models.StationmetaRole(**r)
            db_role.role = get_value_from_str(RC_enum,db_role.role)
            db_role.status = get_value_from_str(RS_enum,db_role.status)
            # check whether role is already present in database
            db_object = get_unique_stationmeta_role(db, db_role.role, db_role.contact_id, db_role.status)
            if db_object:
                role_id = db_object.id
            else:
                db.add(db_role)
                db.commit()
                db.refresh(db_role)
                role_id = db_role.id
            db.execute(insert(stationmeta_core_stationmeta_roles_table).values(station_id=station_id, role_id=role_id))
            db.commit()
    # store annotations and update association table
    if annotations_data:
        # prepare changelog entry/entries
        description = description + f"; add annotation"
        db_old_annotations = get_stationmeta_annotations(db, station_id)
        for olda in db_old_annotations:
            old_annotation = get_stationmeta_annotation_by_id(db, olda.annotation_id)
            old_value=old_value + "{'kind': '" + get_str_from_value(RC_enum,old_role.kind) + \
                                "', 'text': '" + get_str_from_value(RS_enum,old_role.status) + \
                                "', 'date_added': '" + get_str_from_value(RS_enum,old_role.status) + \
                                "', 'approved': '" + get_str_from_value(RS_enum,old_role.status) + \
                                "', 'contributor_id': " + str(old_role.contact_id) + '}'
        if (number_of_commas == 0):
            old_value=old_value + '}'
        else:
            old_value=old_value + ','
        number_of_commas = number_of_commas - 1
        for a in annotations_data:
            db_annotation = models.StationmetaAnnotation(**a)
            # check whether annotation is already present in database
            db_object = get_unique_stationmeta_annotation(db, db_annotation.text, db_annotation.contributor_id)
            if db_object:
                annotation_id = db_object.id
            else:
                db.add(db_annotation)
                db.commit()
                db.refresh(db_annotation)
                annotation_id = db_annotation.id
            db.execute(insert(stationmeta_core_stationmeta_annotations_table).values(station_id=station_id, annotation_id=annotation_id))
            db.commit()
    # store aux_images
    if aux_images_data:
        for i in aux_images_data:
            db_aux_image = models.StationmetaAuxImage(**i)
            db_aux_image.station_id = station_id
            db.add(db_aux_image)
            db.commit()
            db.refresh(db_aux_image)
    # store aux_docs
    if aux_docs_data:
        for d in aux_docs_data:
            db_aux_doc = models.StationmetaAuxDoc(**d)
            db_aux_doc.station_id = station_id
            db.add(db_aux_doc)
            db.commit()
            db.refresh(db_aux_doc)
    # store aux_urls
    if aux_urls_data:
        for u in aux_urls_data:
            db_aux_url = models.StationmetaAuxUrl(**u)
            db_aux_url.station_id = station_id
            db.add(db_aux_url)
            db.commit()
            db.refresh(db_aux_url)
    # store globalmeta
    if globalmeta_data:
        db_global = models.StationmetaGlobal(**globalmeta_data)
        if db_global.climatic_zone:
            db_global.climatic_zone = get_value_from_str(CZ_enum,db_global.climatic_zone)
        db_global.station_id = station_id
        db.add(db_global)
        db.commit()
        db.refresh(db_global)
    # store globalservice
    if globalservice_data:
        db_globalservice = models.StationmetaGlobalService(**globalservice)
        db_globalservice.station_id = station_id
        db.add(db_globalservice)
        db.commit()
        db.refresh(db_globalservice)
    db_changelog = StationmetaChangelog(description=description, station_id=station_id, author_id=1, type_of_change=type_of_change,
                                        old_value=old_value, new_value=str(stationmeta_dict2))
    db.add(db_changelog)
    db.commit()
    # there's a mismatch with coordinates --> how to automatically switch back and forth?!
    db_stationmeta.coordinates = tmp_coordinates
    return db_stationmeta


def delete_stationmeta_field(db: Session, station_id: int, field: str):
    # id can never be deleted (and of course also not changed)!!!
    # there are mandatory fields (from stationmeta_core), that cannot be deleted!
    # --> set these to their default value
    field_table = {'roles': stationmeta_core_stationmeta_roles_table,
                   'annotations': stationmeta_core_stationmeta_annotations_table}
#                  'aux_images': stationmeta_core_stationmeta_aux_images_table,
#                  'aux_docs': stationmeta_core_stationmeta_aux_docs_table,
#                  'aux_urls': stationmeta_core_stationmeta_aux_urls_table,
#                  'globalmeta': stationmeta_core_stationmeta_globalmeta_table}
    if ((field == 'roles') or (field == 'annotations')):
        db.execute(delete(field_table[field]).where(field_table[field].c.station_id==station_id))
    # problem with automatic conversion of coordinates (although not explicitly fetched from database)
    # ==> next two lines are a workaround
    db_stationmeta = db.query(models.StationmetaCore).get(station_id)
    tmp_coordinates = db_stationmeta.coordinates
    db_stationmeta.coordinates = get_geom_from_coordinates(db_stationmeta.coordinates)
    # also write changelog entry
    type_of_change = get_value_from_str(CL_enum,"SingleValue")
    description=f"delete field {field}"
    old_value = get_field_from_record(db, station_id, field, db_stationmeta)
    db_changelog = StationmetaChangelog(description=description, station_id=station_id, author_id=1, type_of_change=type_of_change,
                                        old_value=old_value, new_value="")
    db.add(db_changelog)
    db.commit()
    # there's a mismatch with coordinates --> how to automatically switch back and forth?!
    db_stationmeta.coordinates = tmp_coordinates
    return db_stationmeta
