# -*- coding: utf-8 -*-
"""
Pydantic schemas for TOAR database

"""

from typing import List, Dict
from pydantic import BaseModel, Field, BaseConfig, Json, validator
from geoalchemy2 import WKTElement
from geoalchemy2.shape import to_shape
import datetime as dt
from .models import CZ_enum, CV_enum, ST_enum, TA_enum, TC_enum, \
                    TR_enum, DL_enum, CN_enum, TZ_enum
from toardb.generic.models import RC_enum, RS_enum, AK_enum, CL_enum
from toardb.contacts.schemas import Contact


# the following class was taken from:
# https://github.com/tiangolo/fastapi/issues/312
class Coordinates(BaseModel):
    lat: float = Field(0, gte=-90, lte=90, description="longitude coordinate of station (decimal degrees_east). This is our best estimate of the station location which is not always identical to the official station coordinates (see potential changelog entry).")
    lng: float = Field(0, gte=-180, lte=180, description="latitude coordinate of station (decimal degrees_north). This is our best estimate of the station location which is not always identical to the official station coordinates (see potential changelog entry).")
    alt: float = Field(0, description="altitude of station (in m above sea level). This is our best estimate of the station altitude, which is not always identical to the reported station altitude, but frequently uses the elevation from google earth instead (see potential changelog entry).")

# ======== StationmetaCore =========

class StationmetaCoreBase(BaseModel):
    id: int = Field(None, description="for internal use only")
    codes: List[str] = Field(..., description="list of station's codes")
    name: str = Field(..., description="Station name for use in TOAR; normally equal to data provider’s name")
    coordinates: Coordinates = Field(..., description="Best estimate of station latitude, longitude and altitude")
    coordinate_validation_status: str = Field(..., description="Flag indicating whether the location of a station has been verified")
    country: str = Field(..., description="The country, where the station resides, or which operates the station (e.g. in Antarctica)")
    state: str = Field(..., description="The state or province, where the station resides")
    type_of_environment: str = Field(..., description="Type of station")
    type_of_area: str = Field(..., description="Type of station area")
    timezone: str = Field(..., description="Station timezone")
    additional_metadata: Json = Field(..., description="Any additional metadata about the station as a JSON data structure")

    class Config(BaseConfig):
        orm_mode = True

    @validator('coordinate_validation_status')
    def check_coordinate_validation_status(cls, v):
        return tuple(filter(lambda x: x.value == int(v), CV_enum))[0].string

    @validator('country')
    def check_country(cls, v):
        return tuple(filter(lambda x: x.value == int(v), CN_enum))[0].display_str

    @validator('type_of_environment')
    def check_type_of_environment(cls, v):
        return tuple(filter(lambda x: x.value == int(v), ST_enum))[0].string

    @validator('type_of_area')
    def check_type_of_area(cls, v):
        return tuple(filter(lambda x: x.value == int(v), TA_enum))[0].string

    @validator('timezone')
    def check_timezone(cls, v):
        return tuple(filter(lambda x: x.value == int(v), TZ_enum))[0].string


class StationmetaCorePatch(BaseModel):
    codes: List[str] = None
    name: str = None
    coordinates: Coordinates = None
    coordinate_validation_status: str = None
    country: str = None
    state: str = None
    type_of_environment: str = None
    type_of_area: str = None
    timezone: str = None
    additional_metadata: Json = None

    class Config(BaseConfig):
        orm_mode = True

    @validator('coordinate_validation_status')
    def check_coordinate_validation_status(cls, v):
        if v:
            return tuple(filter(lambda x: x.value == int(v), CV_enum))[0].string

    @validator('country')
    def check_country(cls, v):
        if v:
            return tuple(filter(lambda x: x.value == int(v), CN_enum))[0].string

    @validator('type_of_environment')
    def check_type_of_environment(cls, v):
        if v:
            return tuple(filter(lambda x: x.value == int(v), ST_enum))[0].string

    @validator('type_of_area')
    def check_type_of_area(cls, v):
        if v:
            return tuple(filter(lambda x: x.value == int(v), TA_enum))[0].string

    @validator('timezone')
    def check_timezone(cls, v):
        if v:
            return tuple(filter(lambda x: x.value == int(v), TZ_enum))[0].string


class StationmetaCoreCreate(StationmetaCoreBase):
    pass

    @validator('coordinate_validation_status')
    def check_coordinate_validation_status(cls, v):
        if tuple(filter(lambda x: x.string == v, CV_enum)):
            return v
        else:
            raise ValueError(f"coordinate validation status not known: {v}")

    @validator('country')
    def check_country(cls, v):
        if tuple(filter(lambda x: x.string == v, ST_enum)):
            return v
        else:
            raise ValueError(f"country not known: {v}")

    @validator('type_of_environment')
    def check_type_of_environment(cls, v):
        if tuple(filter(lambda x: x.string == v, ST_enum)):
            return v
        else:
            raise ValueError(f"type of environment not known: {v}")

    @validator('type_of_area')
    def check_type_of_area(cls, v):
        if tuple(filter(lambda x: x.string == v, TA_enum)):
            return v
        else:
            raise ValueError(f"type of area not known: {v}")

    @validator('timezone')
    def check_timezone(cls, v):
        if tuple(filter(lambda x: x.string == v, TZ_enum)):
            return v
        else:
            raise ValueError(f"timezone not known: {v}")


class StationmetaCore(StationmetaCoreBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True


def get_geom_from_coordinates(coordinates: Coordinates) -> str:
    geom_wkte = f'SRID=4326;POINTZ({coordinates.lng} {coordinates.lat} {coordinates.alt})'
    return geom_wkte


def get_coordinates_from_geom(geom: WKTElement) -> Coordinates:
    shply_geom = to_shape(geom)
    coordinates = Coordinates(lng=shply_geom.x, lat=shply_geom.y, alt=shply_geom.z)
    return coordinates

# ======== StationmetaAnnotation =========

class StationmetaAnnotationBase(BaseModel):
    id: int = Field(None, description="for internal use only")
    kind: int = Field(..., description="kind of annotation (see controlled vocabulary: Kind Of Annotations)")
    text: str = Field(..., description="text of annotation")
    date_added: dt.datetime = Field(..., description="timestamp when annotation was added")
    approved: bool = Field(..., description="Flag indicating whether the annotation of a station has been verified")
    contributor_id: int = Field(..., description="ID of contributor who added the annotation")

    @validator('kind')
    def check_kind(cls, v):
        return tuple(filter(lambda x: x.value == int(v), AK_enum))[0].string


class StationmetaAnnotationPatch(BaseModel):
    kind: int = None
    text: str = None
    date_added: dt.datetime = None
    approved: bool = None
    contributor_id: int = None

    @validator('kind')
    def check_kind(cls, v):
        return tuple(filter(lambda x: x.value == int(v), AK_enum))[0].string


class StationmetaAnnotationCreate(StationmetaAnnotationBase):
    pass


class StationmetaAnnotation(StationmetaAnnotationBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True

# ======== StationmetaAux =========

# -------- StationmetaAuxDoc ---------

class StationmetaAuxDocBase(BaseModel):
    id: int = None
    resource_description: str = Field(..., description="Description of the resource")
    date_added: dt.datetime = Field(..., description="Date of metadata record entry into TOAR database")
    resource: str = Field(..., description="auxillary document file")
    station_id: int = Field(..., description="internal station_id to which this resource belongs")


class StationmetaAuxDocPatch(BaseModel):
    resource_description: str = None
    date_added: dt.datetime = None
    resource: str = None
    station_id: int = None


class StationmetaAuxDocCreate(StationmetaAuxDocBase):
    pass


class StationmetaAuxDoc(StationmetaAuxDocBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True

# -------- StationmetaAuxImage ---------

class StationmetaAuxImageBase(BaseModel):
    id: int = None
    resource_description: str = Field(..., description="Description of the resource")
    date_added: dt.datetime = Field(..., description="Date of metadata record entry into TOAR database")
    resource: str = Field(..., description="auxillary image file")
    image_height: int
    image_width: int
    station_id: int = Field(..., description="internal station_id to which this resource belongs")


class StationmetaAuxImagePatch(BaseModel):
    resource_description: str = None
    date_added: dt.datetime = None
    resource: str = None
    image_height: int = None
    image_width: int = None
    station_id: int = None


class StationmetaAuxImageCreate(StationmetaAuxImageBase):
    pass


class StationmetaAuxImage(StationmetaAuxImageBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True

# -------- StationmetaAuxUrl ---------

class StationmetaAuxUrlBase(BaseModel):
    id: int = None
    resource_description: str = Field(..., description="Description of the resource")
    date_added: dt.datetime = Field(..., description="Date of metadata record entry into TOAR database")
    resource: str = Field(..., description="URL")
    station_id: int = Field(..., description="internal station_id to which this resource belongs")


class StationmetaAuxUrlPatch(BaseModel):
    resource_description: str = None
    date_added: dt.datetime = None
    resource: str = None
    station_id: int = None


class StationmetaAuxUrlCreate(StationmetaAuxUrlBase):
    pass


class StationmetaAuxUrl(StationmetaAuxUrlBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True

# ======== StationmetaGlobal =========

class StationmetaGlobalBase(BaseModel):
    population_density_year2010: float = Field(..., description="human population per square km for the year 2010")
    max_population_density_25km_year2010: float = Field(..., description="Year 2010 maximum population density within a radius of 25 km around station location (residents km-2)")
    climatic_zone: str = Field(..., description="Climatic zone according to IPCC, 2006. Data from European Soil Data Centre at JRC, Ispra, Italy")
    nightlight_1km_year2013: float = Field(..., description="Year 2013 nighttime lights brightness values at station location (original 1 km horizontal resolution). Reference: NOAA DMSP stable nighttime lights: a 0.925 km resolution dataset of non-ephemeral visible light intensity at night (Dataset: http://ngdc.noaa.gov/eog/dmsp/downloadV4composites.html)")
    nightlight_5km_year2013: float = Field(..., description="Nighttime lights brightness values for the year 2013 at 5 km resolution (original 1 km horizontal resolution)")
    max_nightlight_25km_year2013: float = Field(..., description="Maximum nighttime lights brightness values for the year 2013 at 25 km resolution (original 1 km horizontal resolution)")
    wheat_production_year2000: float = Field(..., description="Wheat production year2000. Y2000 wheat production amount at station location (units: thousand tons). (Reference: Dataset: http://gaez.fao.org/Main.html)")
    rice_production_year2000: float = Field(..., description="Y2000 rice production amount at station location (units: thousand tons) (Reference: Dataset: http://gaez.fao.org/Main.html)")
    edgar_htap_v2_nox_emissions_year2010: float = Field(..., description="EDGAR emissions inventory for nitrogen oxides (NOx) at 0.1° resolution (Reference: Janssens-Maenhout et al., 2015: http://edgar.jrc.ec.europa.eu/htap_v2/index.php?SECURE=123)")
    omi_no2_column_years2011to2015: float = Field(..., description="Average Y2011-Y2015 tropospheric NO2 columns from OMI at 0.1 degree resolution in units of 10^15 molecules cm-2 (Reference: Krotkov et al., 2016: (no public download))")
    htap_region_tier1: str = Field(..., description="The integer denoting the 'tier1' region defined in the task force on hemispheric transport of air pollution (TFHTAP) coordinated model studies")
    etopo_alt: float = Field(..., description="Global digital elevation model from diverse global and regional datasets in 1 arc minute resolution (Reference: Amante and Eakins, 2009: https://www.ngdc.noaa.gov/mgg/global/global.html)")
    etopo_min_alt_5km: float  = Field(..., description="Minimum elevation from the ETOPO1 dataset in an area of 5 km radius around the station location. This can be used to find out if a high altitude station is located in mountaineous terrain or on a plateau (see etopo_relative_alt)")
    etopo_relative_alt: float = Field(..., description="Station elevation above the surrounding area. Derived by subtracting the minimum altitude within a 5 km radius around the station location from the actual station altitude")
    dominant_landcover_year2012: str = Field(..., description="The dominant IGBP landcover classification at the station location extracted from the MODIS MCD12C1 dataset (3 arc minutes resolution of the year 2012). (Dataset: https://lpdaac.usgs.gov/dataset_discovery/modis/modis_products_table/mcd12c1)")
    toar1_category: str = Field(..., description="The station classification for the Tropsopheric Ozone Assessment Report based on the station proxy data that are stored in the TOAR database")
    station_id: int = Field(..., description="internal station_id to which these global data belong")

    @validator('climatic_zone')
    def check_climatic_zone(cls, v):
        return tuple(filter(lambda x: x.value == int(v), CZ_enum))[0].string

    @validator('toar1_category')
    def check_toar1_category(cls, v):
        return tuple(filter(lambda x: x.value == int(v), TC_enum))[0].string

    @validator('htap_region_tier1')
    def check_htap_region_tier1(cls, v):
        return tuple(filter(lambda x: x.value == int(v), TR_enum))[0].string

    @validator('dominant_landcover_year2012')
    def check_dominant_landcover_year2012(cls, v):
        return tuple(filter(lambda x: x.value == int(v), DL_enum))[0].string


class StationmetaGlobalPatch(BaseModel):
    population_density_year2010: float = None
    max_population_density_25km_year2010: float = None
    climatic_zone: str = None
    nightlight_1km_year2013: float = None
    nightlight_5km_year2013: float = None
    max_nightlight_25km_year2013: float = None
    wheat_production_year2000: float = None
    rice_production_year2000: float = None
    edgar_htap_v2_nox_emissions_year2010: float = None
    omi_no2_column_years2011to2015: float = None
    htap_region_tier1: str = None
    etopo_alt: float = None
    etopo_min_alt_5km: float  = None
    etopo_relative_alt: float = None
    dominant_landcover_year2012: str = None
    toar1_category: str = None
    station_id: int = None

    @validator('climatic_zone')
    def check_climatic_zone(cls, v):
        return tuple(filter(lambda x: x.value == int(v), CZ_enum))[0].string

    @validator('toar1_category')
    def check_toar1_category(cls, v):
        return tuple(filter(lambda x: x.value == int(v), TC_enum))[0].string

    @validator('htap_region_tier1')
    def check_htap_region_tier1(cls, v):
        return tuple(filter(lambda x: x.value == int(v), TR_enum))[0].string

    @validator('dominant_landcover_year2012')
    def check_dominant_landcover_year2012(cls, v):
        return tuple(filter(lambda x: x.value == int(v), DL_enum))[0].string


class StationmetaGlobalCreate(StationmetaGlobalBase):
    pass

    @validator('climatic_zone')
    def check_climatic_zone(cls, v):
        if tuple(filter(lambda x: x.string == v, CZ_enum)):
            return v
        else:
            raise ValueError(f"climatic zone not known: {v}")

    @validator('toar1_category')
    def check_toar1_category(cls, v):
        if tuple(filter(lambda x: x.string == v, TC_enum)):
            return v
        else:
            raise ValueError(f"TOAR1 category not known: {v}")

    @validator('htap_region_tier1')
    def check_htap_region_tier1(cls, v):
        if tuple(filter(lambda x: x.string == v, TR_enum)):
            return v
        else:
            raise ValueError(f"HTAP region TIER1 not known: {v}")

    @validator('dominant_landcover_year2012')
    def check_dominant_landcover_year2012(cls, v):
        if tuple(filter(lambda x: x.string == v, DL_enum)):
            return v
        else:
            raise ValueError(f"dominant landcover (year2012) not known: {v}")


class StationmetaGlobal(StationmetaGlobalBase):
    class Config:
        orm_mode = True


class StationmetaGlobalBaseNested(BaseModel):
    population_density_year2010: float = None
    max_population_density_25km_year2010: float = None
    climatic_zone: str = None
    nightlight_1km_year2013: float = None
    nightlight_5km_year2013: float = None
    max_nightlight_25km_year2013: float = None
    wheat_production_year2000: float = None
    rice_production_year2000: float = None
    edgar_htap_v2_nox_emissions_year2010: float = None
    omi_no2_column_years2011to2015: float = None
    htap_region_tier1: int = None
    etopo_alt: float = None
    etopo_min_alt_5km: float  = None
    etopo_relative_alt: float = None
    dominant_landcover_year2012: int = None
    toar1_category: str = None


class StationmetaGlobalNestedCreate(StationmetaGlobalBaseNested):
    pass


class StationmetaGlobalNested(StationmetaGlobalBaseNested):
    class Config:
        orm_mode = True


# ======== StationmetaGlobalService =========

class StationmetaGlobalServiceBase(BaseModel):
    id: int = None
    variable_name: str = Field(..., description="Variable name which will be returned by the service URL")
    result_type: int = Field(..., description="Result type")
    result_nvalues: int = Field(..., description="Expected number of return values (1-16)")
    service_valid_year: int = Field(..., description="Year for which metadata are returned from this service URL or null if undefined")
    service_url: str = Field(..., description="Service URL with %LAT% and %LON% placeholders to enter station or point location in floating point degrees")


class StationmetaGlobalServiceCreate(StationmetaGlobalServiceBase):
    pass


class StationmetaGlobalService(StationmetaGlobalServiceBase):
    id: int = Field(..., description="for internal use only")

    class Config:
        orm_mode = True

# ======== StationmetaRole =========

class StationmetaRoleBase(BaseModel):
    id: int = None
    role: str = Field(..., description="Role of contact (see controlled vocabulary: Role Codes)")
    status: str = Field(..., description="Status of contact (see controlled vocabulary: Role Status)")
#   contact: Contact
#   at the moment contact_id has to be given...
#   ==> in the future: give unique contact_email
#   patching stationmeta should not result in creating new contacts!
#   ==> still to do: check, whether contact already exists (otherwise patching cannot be done)
    contact_id: int = Field(..., description="to be deleted")

    @validator('role')
    def check_role(cls, v):
        return tuple(filter(lambda x: x.value == int(v), RC_enum))[0].string

    @validator('status')
    def check_status(cls, v):
        return tuple(filter(lambda x: x.value == int(v), RS_enum))[0].string

    class Config:
        orm_mode = True


class StationmetaRolePatch(BaseModel):
    role: str
    status: str
#   at the moment contact_id has to be given...
#   ==> in the future: give unique contact_email
#   patching stationmeta should not result in creating new contacts!
#   ==> still to do: check, whether contact already exists (otherwise patching cannot be done)
    contact_id: int

    class Config:
        orm_mode = True


class StationmetaRoleCreate(StationmetaRoleBase):
    pass

    @validator('role')
    def check_role(cls, v):
        if tuple(filter(lambda x: x.string == v, RC_enum)):
            return v
        else:
            raise ValueError(f"role code not known: {v}")

    @validator('status')
    def check_status(cls, v):
        if tuple(filter(lambda x: x.string == v, RS_enum)):
            return v
        else:
            raise ValueError(f"role status not known: {v}")


class StationmetaRole(StationmetaRoleBase):
    id: int = Field(..., description="for internal use only")
    contact: Contact = Field(..., description="Contact for this role")

    class Config:
        orm_mode = True


# ======== StationmetaChangelog =========

class StationmetaChangelogBase(BaseModel):
    datetime: dt.datetime = Field(..., description="Date of change to the TOAR database")
    description: str = Field(..., description="Description of change")
    old_value: str = Field(..., description="value that has been changed in the TOAR database")
    new_value: str = Field(..., description="new, changed value")
    station_id: int = Field(..., description="internal ID of station to which this change belongs")
    author_id: int = Field(..., description="internal ID of person who submitted the change to the TOAR database")
    type_of_change: int = Field(..., description="Type of change (see controlled vocabulary: Type Of Change)")


    @validator('type_of_change')
    def check_role(cls, v):
        return tuple(filter(lambda x: x.value == int(v), CL_enum))[0].string

    class Config:
        orm_mode = True

class StationmetaChangelog(StationmetaChangelogBase):

    class Config:
        orm_mode = True


# ======== for nested view/upload =========

class StationmetaBase(StationmetaCoreBase):
    roles: List[StationmetaRole] = None
    annotations: List[StationmetaAnnotation] = None
    aux_images: List[StationmetaAuxImage] = None
    aux_docs: List[StationmetaAuxDoc] = None
    aux_urls: List[StationmetaAuxUrl] = None
    globalmeta: StationmetaGlobal = None
    changelog: List[StationmetaChangelog] = None

    class Config:
       orm_mode = True


class StationmetaPatch(StationmetaCorePatch):
    roles: List[StationmetaRolePatch] = None
    annotations: List[StationmetaAnnotationPatch] = None
    aux_images: List[StationmetaAuxImagePatch] = None
    aux_docs: List[StationmetaAuxDocPatch] = None
    aux_urls: List[StationmetaAuxUrlPatch] = None
    globalmeta: StationmetaGlobalPatch = None

    class Config:
       orm_mode = True


class StationmetaCreate(StationmetaCoreCreate):
    roles: List[StationmetaRoleCreate] = None
    annotations: List[StationmetaAnnotation] = None
    aux_images: List[StationmetaAuxImage] = None
    aux_docs: List[StationmetaAuxDoc] = None
    aux_urls: List[StationmetaAuxUrl] = None
    globalmeta: StationmetaGlobalNestedCreate = None

    class Config:
        orm_mode = True


class Stationmeta(StationmetaBase):
    id: int = Field(..., description="for internal use only")

    class Config:
      orm_mode = True

